from __future__ import print_function
import pickle
import warnings
from os.path import isfile, join

from AutoRPE.UtilsRPE.BasicStructures import Interface, Module, Routine, Function
from AutoRPE.UtilsRPE.Vault import Vault
from AutoRPE.UtilsRPE.ObtainSourceFileInfo import ObtainSourcefileInfo
from AutoRPE.UtilsRPE.CurrentBlock import CurrentBlock
from AutoRPE.UtilsRPE.Counter import Counter
import AutoRPE.UtilsRPE.Error as Error
import AutoRPE.UtilsRPE.PointerDealer as PointerDealer
from AutoRPE.UtilsRPE.TypeDealer import DerivedType
import AutoRPE.UtilsRPE.VariablePrecision as VariablePrecision
from AutoRPE.UtilsRPE.Functions import insert_variable_precision_specification_to_allocatable, \
    insert_variable_precision_specification_to_namelist_parameter, insert_variable_precision_to_routine_variable, \
    load_source_files, listdir_ord, split_vars_by_allocatable, split_vars_by_main_or_routine, \
    split_by_appearance_in_namelist, calls_to_routines, close_brackets, find_called_routine, find_call_arguments, \
    is_keyword_argument, match_arguments_and_interface, is_call_to_function, has_operations, type_of_contents, \
    insert_in_line


def parse_sources(path_to_model_sources, save_vault=False):
    # Find the nemo source-files in sources
    # (Using the sources that already modified replacing real declarations by type(rpe_var) )
    sources = [f for f in listdir_ord(path_to_model_sources) if
               isfile(join(path_to_model_sources, f)) and f.count(".f90")]
    # Initialize two empty dictionaries that will contain all the routines and all the functions
    object_oriented_dictionary = {"procedures": {}, "interfaces": {}, "variables": {}, "modules": {},
                                  "derivedtypes": {}}

    # Initialize an empty dictionary that will contain all the source-files.
    sourcefiles = {}
    vault = Vault()

    print("\n\nParsing sourcecode...\n")

    # Going over all the source-files loading the files,
    # parsing the routines and storing both things in dictionaries.

    for index, filename in enumerate(sources):
        module_name = filename.replace(".f90", "")
        sourcefiles[module_name] = ObtainSourcefileInfo(join(path_to_model_sources, filename))
        module = Module(module_name)
        f = sourcefiles[module_name]
        f.module = module
        f.find_routines()
        f.find_functions()
        f.find_module_variables()
        for procedure in f.objects:
            object_oriented_dictionary["procedures"][procedure.name] = procedure
            module.add_procedure(procedure)
            if procedure.name == "main":
                module.variables = procedure.variables
                for var in procedure.variables:
                    vault.variables.append(var)
        vault.add_module(module)
        for procedure in f.objects:
            vault.procedures.append(procedure)
            if isinstance(procedure, Routine):
                vault.routines.append(procedure)
                for var in procedure.variables:
                    vault.variables.append(var)
            if isinstance(procedure, Function):
                vault.functions.append(procedure)
                for var in procedure.variables:
                    vault.variables.append(var)


    # Going over all the source-files parsing the interfaces and storing them in dictionary.
    for index, filename in enumerate(sources):
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_interfaces(object_oriented_dictionary["procedures"])
        # Getting and saving modules used
        f.module = vault[module_name]
        f.find_used_modules(vault)
        for interface in f.objects:
            if isinstance(interface, Interface):
                object_oriented_dictionary["interfaces"][interface.name] = interface
                vault[module_name].procedures.append(interface)
                vault.interfaces.append(interface)
                vault.procedures.append(interface)

    # After having all the used parsed, propagate the inheritance, and add public var to available vars
    vault.expand_use_statement()

    # Going over all the source-files parsing the derived types and storing them in dictionary.
    for index, filename in enumerate(sources):
        module_name = filename.replace(".f90", "")
        f = sourcefiles[module_name]
        f.find_types()
        for derived_type in f.objects:
            if isinstance(derived_type, DerivedType):
                object_oriented_dictionary["derivedtypes"][derived_type.name] = derived_type
                vault[module_name].procedures.append(derived_type)
                vault.derived_types.append(derived_type)
                vault.procedures.append(derived_type)
                for var in derived_type.variables:
                    vault.variables.append(var)

    # Fill a dictionary to speed up search of variables
    vault.process_variables()
    # Now that vault.derived_types is full add extension to derived types if any
    vault.expand_derived_types()

    # Passing to pointed objects the actual function/subroutine they points to
    PointerDealer.process_pointers(vault)

    # # Find implicit function declarations
    # for filename in sourcefiles:
    #     module_name = filename.replace(".f90", "")
    #     f = sourcefiles[module_name]
    #     f.find_implicit_functions(vault=vault)
    #     for procedure in f.objects:
    #         if procedure not in vault[module_name].procedures:
    #             vault[module_name].procedures.append(procedure)
    #             vault.procedures.append(procedure)

    print("number of functions " + str(len(vault.functions)) + "\n" +
          "number or routines " + str(len(vault.routines)) + "\n" +
          "number or interfaces " + str(len(vault.interfaces)) + "\n" +
          "number or main " + str(len(vault.modules)) + "\n" +
          "number or procedures " + str(len(vault.procedures)) + "\n" +
          "number of variables " + str(len(vault.variables)) + "\n" +
          "number or derived_types " + str(len(vault.derived_types)) + "\n")

    if save_vault:
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))
    return vault


def preprocess_sources(path_to_input_sources, path_to_output_sources, blacklist=[]):
    from AutoRPE.UtilsRPE.ImplementRPE import ImplementRPE
    # Read all files sources excluding hidden files ( like vim swp files)
    all_sources = [f for f in listdir_ord(path_to_input_sources) if
                   isfile(join(path_to_input_sources, f)) and f.count(".f90") and f[0][0] != "."]
    total = len(all_sources)
    if len(all_sources):
        # Load and format files
        for index, filename in enumerate(all_sources):
            f = ImplementRPE(join(path_to_input_sources, filename), None)
            # Remove comments, empty lines and old style declarations
            if filename in blacklist:
                print("\rFormatting file %i/%i  %20s" % (index + 1, total, filename), end="")
                f.pre_process()
            # Additionally reals are replaced with rpe_var type
            else:
                print("\rFormatting file and replacing reals %i/%i  %20s" % (index + 1, total, filename), end="")
                f.pre_process(replace_reals=True)
            f.save_modified_file(join(path_to_output_sources, filename))
    else:
        # Sources is an empty folder
        raise ValueError("No files found in %s folder", path_to_input_sources)


def postprocess_sources(path_to_input_sources, path_to_output_sources, vault):
    from AutoRPE.UtilsRPE.ImplementRPE import ImplementRPE

    # List files in source sexcluding hidden files ( like vim swp files)
    all_sources = [f for f in listdir_ord(path_to_input_sources) if
                   isfile(join(path_to_input_sources, f)) and f.count(".f90")and f[0][0] != "."]
    total = len(all_sources)
    for index, filename in enumerate(all_sources):
        f = ImplementRPE(join(path_to_input_sources, filename), vault)
        print("\rFixing issues %i/%i  %20s" % (index + 1, total, filename), end="")
        f.post_process()
        f.save_modified_file(join(path_to_output_sources, filename))


def assign_sbits_to_rpe_variables(path_to_sources, vault=None):
    print("\n\nAssigning significant bits...")

    def condition(v):
        return v.type == "rpe_var" and not v.is_parameter and not v.name.count("tmp_var_")

    source_files = load_source_files(path_to_sources)
    rpe_variables = [var for var in vault.variables if condition(var) and not var.dimension == None]
    allocatable, non_allocatable = split_vars_by_allocatable(rpe_variables)
    var_counter = Counter()
    # Allocatables:
    for variable in allocatable:
        insert_variable_precision_specification_to_allocatable(variable, source_files, var_counter)

    # Non allocatables that appear in a namelist:
    na_in_namelist, na_not_in_namelist = split_by_appearance_in_namelist(non_allocatable, source_files)
    for variable in na_in_namelist:
        insert_variable_precision_specification_to_namelist_parameter(variable, source_files, var_counter)
    in_main, not_in_main = split_vars_by_main_or_routine(na_not_in_namelist)

    # Variables inside routines
    for variable in not_in_main:
        if variable.intent != "in":
            insert_variable_precision_to_routine_variable(variable, source_files, var_counter)

    # Adding sbits assignation and apply_truncation after routine calls
    insert_variable_precision_after_routine_call(source_files, vault=vault, counter=var_counter)

    # To save the variable id into the vault.
    save = True
    if save:
        # Saving modified files
        for filename, source_file in source_files.items():
            source_file.save_modified_file(join(path_to_sources, filename))
        # This workaround is needed to allow pickle to write the file, don't know yet about its risks
        import sys
        sys.setrecursionlimit(2000)
        pickle.dump(vault, open("vault.pkl", "wb"))
    return


def create_read_precisions_module_and_add_to_file(path_to_sources, target_module=None, vault=None):
    from os.path import dirname, abspath
    print("\n\nCreating module to read precision namelist...")
    # Adding the use statement in all the files.
    line_to_add = "   USE read_precisions\n"
    source_files = load_source_files(path_to_sources)

    for name, source_file in source_files.items():
        source_file.lines.insert(1, line_to_add)
        source_file.rebuild_text()
        source_file.save_modified_file(join(path_to_sources, source_file.filename))

    # Create module
    rpe_vars = [var for var in vault.variables if var.type == "rpe_var"]
    assigned_rpe_vars = [var for var in rpe_vars if var.mutable]
    ids = [var.id for var in assigned_rpe_vars]
    # Path to template is hardcoded
    package_directory = dirname(abspath(__file__))
    filename = package_directory + "/../AdditionalFiles/read_precisions_template"

    with open(filename, "r") as input_file:
        template = input_file.read()
    module = template.replace("%NUMBER_OF_RPE_VARS%", "%i" % max(ids))
    with open(join(path_to_sources, "read_precisions.f90"), "w") as output_file:
        output_file.write(module)
    # Put call into target module

    if target_module:
        if isfile(join(path_to_sources, target_module)):
            with open(join(path_to_sources, target_module), "r") as input_file:
                original_source = input_file.read()

            lines = original_source.split("\n")
            lines.insert(1, "use read_precisions")
            for index, line in enumerate(lines):
                if line.lower().strip() == "private":
                    lines.pop(index)
            original_source = "\n".join(lines)

            with open(join(path_to_sources, target_module), "w") as output_file:
                output_file.write(original_source)
        else:
            warnings.warn("File %s not found in sources.\nUnable to insert read_precisions module")


def insert_variable_precision_after_routine_call(source_files, vault, counter):
    for name, source_file in source_files.items():
        try:
            current_block = CurrentBlock(vault[source_file.filename.replace(".f90", "")].main())
        except AttributeError:
            warnings.warn("No module %s" % source_file.filename)
            return

        for _index, _line in enumerate(source_file.lines):
            # Find routine calls in line
            routine_calls = calls_to_routines(_line)
            variables_to_truncate = []
            if not routine_calls:
                # If not, look if the code is entering or exiting a routine block
                current_block = current_block.in_which_block(_line, vault=vault)
            else:
                # Keep the condition if exists
                if _line.lower().strip().find("if(") == 0:
                    condition = close_brackets(_line)

                else:
                    condition = None
                # If there's some call look at each one of them
                for call in routine_calls:
                    is_external = False
                    called_routine = find_called_routine(call)
                    # If the routine is apply_truncation, continue
                    if called_routine == "apply_truncation":
                        continue
                    called_arguments = find_call_arguments(call)
                    if not called_arguments:
                        # If there's no argument, continue
                        continue

                    # Check that we properly found the routine in the database
                    try:
                        # Find the procedure called (object)
                        procedure = vault.get_procedure_by_name(called_routine)
                        if isinstance(procedure, Routine):
                            pass
                        elif isinstance(procedure, Interface):
                            procedure = match_arguments_and_interface(called_arguments, interface=procedure,
                                                                      routine=current_block.routine, vault=vault)
                    except Error.ProcedureNotFound:
                        warnings.warn("Routine not found in vault database %s" % called_routine)
                        is_external = True

                    # If the routine is in the database
                    if not is_external:
                        # Check each argument
                        for argument_index, argument in enumerate(called_arguments):
                            if argument.count("tmp_var_"):
                                # Skipping temporal variables
                                continue
                            # Look if argument is keyword argument
                            if is_keyword_argument(argument):
                                # In case it is a keyword argument, get the dummy argument using the name.
                                keyword = argument.split("=")[0].strip()
                                dummy_argument = procedure.get_variable_by_name(keyword)
                                if not dummy_argument:
                                    raise AssertionError("Keyword variable not found: %s" % keyword)

                                if dummy_argument.type != "rpe_var":
                                    # If the argument is not an rpe_var because it's passing the %val element of the struct
                                    # it should be truncated anyway.
                                    if dummy_argument.type in VariablePrecision.real_types and argument.count("%val"):
                                        variables_to_truncate.append(argument.replace("%val", "").strip())
                                elif dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                    variables_to_truncate.append(argument)
                            else:
                                dummy_argument = procedure.get_variable_by_position(argument_index)
                                if not dummy_argument:
                                    raise AssertionError("Variable not found in position: %i" % argument_index)
                                if dummy_argument.type != "rpe_var":
                                    if dummy_argument.type == VariablePrecision.real_types and argument.count("%val"):
                                        # If the argument contains whichever operations, for sure is intent in
                                        is_c2f = is_call_to_function(argument)
                                        has_op = has_operations(argument)
                                        if not is_c2f and not has_op:
                                            argument = argument.split("%val", 1)[0]
                                            if argument:
                                                variables_to_truncate.append(argument.strip())
                                            else:
                                                raise AssertionError
                                elif dummy_argument.intent == "out" or dummy_argument.intent == "inout":
                                    variables_to_truncate.append(argument)
                    else:
                        # If the routine is external
                        for argument_index, argument in enumerate(called_arguments):
                            # TODO: Probably should find a better solution, the code should not identify an array as a
                            # Todo: ... function even if the name coincides.
                            argument_type = type_of_contents(argument, routine=current_block.routine, vault=vault)
                            if argument_type == "rpe_var":
                                variables_to_truncate.append(argument)
                            elif argument_type in VariablePrecision.real_types and argument.count("%val"):
                                variables_to_truncate.append(argument.replace("%val", "").strip())
            i = _index
            for variable in variables_to_truncate:
                i = insert_in_line(variable, source_file, current_block.routine, counter, i, vault=vault,
                                   condition=condition)

        source_file.rebuild_text()
        return
